structure KSatrak :> KSatrak =
    (* dp01s nhf, v1.2 -- 2001. mj. 5. Hank Pter *)
struct

(* A bemenetet, ill. a kimenetet a terminlra irnythatjuk Unix alatt
   a /dev/stdin, ill. a /dev/stdout, MS-DOS alatt mindkt esetben a con
   llomnynv megadsval.
*)

exception KSatrak of string;

type parcMintak = (TSatrak.parc * string) list;

(* satrak_be f = az f szveges llomnybl beolvasott adatokbl kpzett feladvnyler
   satrak_be : string -> TSatrak.fLeiro
*)
fun satrak_be f =
    let
	(* feladvanymatrix istr zs zss = az istr nyel llomnybl beolvasott feladvnymtrix
	     sorlerjbl, oszloplerjbl s a parcellk listjnak listjbl ll hrmas,
             ahol a zs gyjti az oszlopler els elemt s a sorlert, a zss pedig az
             oszlopler tbbi elemt s a parcellasorokat;
             az res sorokat s a sorokon bell az ismtld formz karaktereket figyelmen
             kvl hagyja
           feladvanymatrix : instream -> string list -> string list list ->
	                                 string list * string list * string list list
	*)
	fun feladvanymatrix istr zs zss =
	      if TextIO.endOfStream istr 
	      then let val (ks, kss) = (rev zs, rev zss)
		   in
		       (tl ks, hd ks :: hd kss, tl kss)
		   end
	      else let val xs = String.tokens Char.isSpace (TextIO.inputLine istr)
		   in
		       if null xs then feladvanymatrix istr zs zss
		       else feladvanymatrix istr (hd xs :: zs) (tl xs :: zss)
		   end

	(* feladvany f = az f nev llomnybl beolvasott feladvnymtrix sorlerjbl,
                         oszloplerjbl s a parcellk listjnak listjbl ll hrmas
           feladvany : string -> string list * string list * string list list
	*)
	fun feladvany f =
	      let val istr = TextIO.openIn f
	      in
		  feladvanymatrix istr [] [] before TextIO.closeIn istr
	      end

	(* fahelyek fss n m zs = a kertben a fkat tartalmaz parcellk sor- s
               oszlopszmnak a listja, ahol 
               fss az sszes parcella listjnak listja, 
               n a sorokat s m az oszlopokat szmllja, 
               zs a fkat tartalmaz parcellk sor- s oszlopszmt gyjti;
               KSatrak nven hibt jelez, ha nem "F" vagy "-" a parcella jele
	   fahelyek : string list list -> int -> int -> TSatrak.parc list -> TSatrak.parc list
	*)
	fun fahelyek (("F"::fs)::fss) n m zs = fahelyek (fs::fss) n (m+1) ((n,m)::zs)
	  | fahelyek (("-"::fs)::fss) n m zs = fahelyek (fs::fss) n (m+1) zs
	  | fahelyek ((_::_)::_) n m zs =
	      raise KSatrak("Ismeretlen jel van a(z) (" ^ Int.toString n ^ ", " ^
			    Int.toString m ^ ") parcelln.\n")
	  | fahelyek ([]::fss) n m zs = fahelyek fss (n+1) 1 zs
	  | fahelyek [] n m zs = rev zs

	val (ss, os, fss) = feladvany f
	val sSzS = List.mapPartial Int.fromString ss
	val sSzO = List.mapPartial Int.fromString os
	val _ = if foldl (fn (fs, b) => length os = length fs andalso b) true fss
		then ()
		else raise KSatrak("Az oszloplerban s a feladvnymtrix egy \
		                   \vagy tbb sorban az elemek szma klnbz.\n")
	val fs = fahelyek fss 1 1 []
    in
	(sSzS, sSzO, fs)
    end;

(* compare (((i, j), _), ((p, r), _)) = (r,p)>(j,i) --> LESS, (r,p)>(j,i) --> GREATER
   compare : TSatrak.parc * 'a -> order
*)
fun compare (((i, j), _), ((p, r), _)) = if p<i orelse p=i andalso r<j then LESS
                                         else if p>i orelse p=i andalso r>j then GREATER
                                         else EQUAL;

(* fakEsSatrak ((n, m), fs, ss) =  az adott parcellasorban a fk helyt ler fs lista s 
          a mr megtallt strak sajt fjukhoz viszonytott helyzett ler ss lista
          felhasznlsval ellltott megjelent-lista a sor- s oszlopszmok cskken
          sorrendjben; 
         (n div 2, m) a jobb als sarokban lv parcella sor- s oszlopszma: az aktulis
           sor- s oszlopszm rvnyesgnek ellenrzsre hasznljuk;
         KSatrak nven hibt jelez, ha a kirand megoldsban a fk s a strak szma nem
         azonos, vagy ha valamelyik sor- s oszlopszm nem ltez parcellra hivatkozik
   fakEsSatrak : (int * int) * TSatrak.parc list * TSatrak.sHelyek -> parcMintak
*)
fun fakEsSatrak ((n, m), fs, ss) =
      if length fs <> length ss 
      then raise KSatrak("A fak es a satrak szama (" ^ Int.toString(length fs) ^
                         " ill. " ^ Int.toString(length ss) ^ ") kulonbozo.")
      else 
	let 
	    (* joParc (i, j) = (i, j), ha az (i, j) helyen van parcella a kertben,
                               egybknt KSatrak nven hibt jelez
	       joParc : TSatrak.parc -> TSatrak.parc
	    *)
	    fun joParc (i, j) =
	      let val k = (i + 1) div 2
	      in
		  if 1 <= k andalso 1 <= j andalso k <= n andalso j <= m 
		      then (i,j)
		  else  raise KSatrak("Nincs parcella a(z) (" ^ Int.toString k ^
				", " ^ Int.toString j ^ ") helyen.")
	      end

	    (* sH (fs, ss, zs) = a fk helyt megad fs lista s a strak sajt fjukhoz
                     viszonytott helyzett ler ss lista alapjn kibvtett zs
                     megjelent-lista 
               sH : TSatrak.parc list * TSatrak.shelyek * parcMintak -> parcMintak
	    *)
	    fun sH ((i, j)::fs, TSatrak.w::ss, zs) =
		    sH (fs, ss, (joParc(2*i-1, j-1), "| S")::(joParc(2*i-1, j), "-F ")::zs)
	      | sH ((i, j)::fs, TSatrak.n::ss, zs) =
		    sH (fs, ss, (joParc(2*i-3, j), "|S ")::(joParc(2*i-2, j), "+|-")
                                                         ::(joParc(2*i-1, j), "|F ")::zs)
	      | sH ((i, j)::fs, TSatrak.e::ss, zs) =
		    sH (fs, ss, (joParc(2*i-1, j), "| F")::(joParc(2*i-1, j+1), "-S ")::zs)
	      | sH ((i, j)::fs, TSatrak.s::ss, zs) =
		    sH (fs, ss, (joParc(2*i-1, j), "|F ")::(joParc(2*i, j), "+|-")
                                                         ::(joParc(2*i+1, j), "|S ")::zs)
	      | sH (_, _, zs) = zs
	in
	    Listsort.sort compare (sH (fs, ss, []))
	end;

(* mRajzok ((n, m), fs, sss) = a jobb als sarokban tallhat (n div 2, m) sor- s oszlopszm
       parcelltl kezdve, htulrl visszafel haladva felpti a megoldsok sss listjt 
       megjelent fzrt
   mRajzok : (int * int) * (int * int) list * TSatrak.helyek list -> string
*)
fun mRajzok (nm as (n, m), fs, ss::sss) =
      let
	  (* pat (i, p1, p2) = p1, ha i pros, egybknt p2
	     pat : int * string * string -> string
	  *)
	  fun pat (i, p1 : string, p2) = if i mod 2 = 0 then p1 else p2

	  (* mRajz ((i, j), qs, t) = a qs listabeli fzrelemek az (i, j)-k cskken 
                           sorrendjben s a megfelel sormintk a t el fzve 
	     mRajz : (int * int) * parcMintak * string -> string
	  *)
	  fun mRajz ((0, 0), qs, t) = t
	    | mRajz ((i, 0), qs, t) = 
                         mRajz((i-1, m), qs, pat(i, "|", "+") ^ "\n" ^ t)
	    | mRajz ((i, j), (qqs as (nm, s)::qs), t) =
		         if (i, j) = nm then mRajz((i, j - 1), qs, (s ^ t))
		         else mRajz((i, j - 1), qqs,  pat(i, "+--", "|  ") ^ t)
	    | mRajz ((i, j), [], t) = 
                         mRajz((i, j - 1), [], pat(i, "+--", "|  ") ^ t)
      in
	  mRajz(nm, fakEsSatrak(nm, fs, ss), "+\n") ^ mRajzok(nm, fs, sss)
      end
  | mRajzok (_, _, []) = "";

(* satrak_ki(m, sss, fl) = az sss megoldslista s az fl feladvnyler
      alapjn az sszes megoldst kirja az m szveges llomnyba
   satrak_ki : string * TSatrak.fLeiro * TSatrak.sHelyek list -> unit
*)
fun satrak_ki (m, (sSzS, sSzO, fs), sss) =
      let val ostr = TextIO.openOut m
	  val _ = TextIO.output(ostr, mRajzok((2 * length sSzS, length sSzO), fs, sss))
      in
           TextIO.closeOut ostr
      end;

(* megold(f, m) = beolvas egy feladvnyt az f szveges llomnybl s kirja
      az sszes megoldst az m szveges llomnyba; ehhez szksge van a
      Satrak.satrak fggvnyre; eredmnye az f nevet, a megoldsok szmt s a
      futsi idt mp-ben tartalmaz fzr
   megold : string * string -> string
*)
fun megold (f, m) =
    let val (sSzS, sSzO, fs) = satrak_be f
	val ostr = TextIO.openOut m
	fun megoldas_ki sss = 
              TextIO.output(ostr, mRajzok((2 * length sSzS, length sSzO), fs, sss))
	val start = Timer.startCPUTimer()
	val sss = Satrak.satrak(sSzS, sSzO, fs)
	val {usr=tim,...} = Timer.checkCPUTimer start
	val _ = megoldas_ki sss
	val _ = TextIO.closeOut ostr
	val t = "Table: " ^ f ^ ", " ^
                Int.toString(length sss) ^ " solution(s), " ^
		"time: " ^ Time.fmt 2 tim ^ " sec\n"
    in
	t
    end

end
